import React, { Component } from 'react';
import { AuctionList } from './AuctionList';
import { AuctionDetails } from './AuctionDetails';
import { NewAuction, NewAuctionInfo } from './NewAuction';
import { Auction, parseAuction } from './auction';
import { isRecord } from './record';


// Indicates which page to show. If it is the details page, the argument
// includes the specific auction to show the details of.
type Page = "list" | "new" | {kind: "details", index: number}; 

// RI: If page is "details", then index is a valid index into auctions array.
type AppState = {page: Page, auctions: ReadonlyArray<Auction> | undefined};

// Whether to show debugging information in the console.
const DEBUG: boolean = true;


// Top-level component that displays the appropriate page.
export class App extends Component<{}, AppState> {

  constructor(props: {}) {
    super(props);
    this.state = {page: "list", auctions: []};
  }

  componentDidMount = (): void => {
    this.doRefreshAllClick();
  };

  render = (): JSX.Element => {
    if (this.state.auctions === undefined) {
      if (DEBUG) console.debug("rendering loading page");
      return <div>Loading...</div>;

    } else if (this.state.page === "list") {
      if (DEBUG) console.debug("rendering list page");
      return <AuctionList auctions={this.state.auctions}
                          onNewClick={this.doNewClick}
                          onAuctionClick={this.doAuctionClick}
                          onRefreshClick={this.doRefreshAllClick}/>;

    } else if (this.state.page === "new") {
      if (DEBUG) console.debug("rendering add page");
      return <NewAuction onStartClick={this.doStartClick}
                         onBackClick={this.doBackClick}/>;

    } else {  // details
      const index = this.state.page.index;
      const auction = this.state.auctions[index];
      if (DEBUG) console.debug(`rendering details page for "${auction.name}"`);
      return <AuctionDetails auction={auction}
                  onBidClick={(bidder, amt) => this.doBidClick(index, bidder, amt)}
                  onRefreshClick={() => this.doRefreshAuctionClick(index)}
                  onBackClick={this.doBackClick}/>;
    }
  };
  
  doRefreshAllClick = (): void => {
    fetch("/api/list").then(this.doListResp)
        .catch(() => this.doListError("failed to connect to server"));
  };

  doListResp = (resp: Response): void => {
    if (resp.status === 200) {
      resp.json().then(this.doListJson)
          .catch(() => this.doListError("200 response is not JSON"));
    } else if (resp.status === 400) {
      resp.json().then(this.doListError)
          .catch(() => this.doListError("400 response is not text"));
    } else {
      this.doListError(`bad status code from /api/list: ${resp.status}`);
    }
  };

  doListJson = (data: unknown): void => {
    if (!isRecord(data)) {
      console.error("bad data from /api/list: not a record", data);
      return;
    }

    if (!Array.isArray(data.auctions)) {
      console.error("bad data from /api/list: auctions is not an array", data);
      return;
    }

    const auctions: Auction[] = [];
    for (const val of data.auctions) {
      const auction = parseAuction(val);
      if (auction === undefined)
        return;
      auctions.push(auction);
    }
    this.setState({auctions});
  };

  doListError = (msg: string): void => {
    console.error(`Error fetching /api/list: ${msg}`);
  };

  doNewClick = (): void => {
    if (DEBUG) console.debug("set state to new");
    this.setState({page: "new"});
  };
  
  doAuctionClick = (index: number): void => {
    if (DEBUG) console.debug(`set state to details for auction ${index}`);
    this.setState({page: {kind: "details", index}});
  };
  
  doBackClick = (): void => {
    if (DEBUG) console.debug("set state to list");
    this.setState({page: "list"});
  };

  doStartClick = (info: NewAuctionInfo): void => {
    if (DEBUG) console.debug("add an auction and then show the full list");

    fetch("/api/add", {
        method: "POST", body: JSON.stringify(info),
        headers: {"Content-Type": "application/json"} })
      .then(this.doAddResp)
      .catch(() => this.doAddError("failed to connect to server"));
  };

  doAddResp = (resp: Response): void => {
    if (resp.status === 200) {
      resp.json().then(this.doAddJson)
          .catch(() => this.doAddError("200 response is not JSON"));
    } else if (resp.status === 400) {
      resp.json().then(this.doAddError)
          .catch(() => this.doAddError("400 response is not text"));
    } else {
      this.doAddError(`bad status code from /api/add: ${resp.status}`);
    }
  };

  doAddJson = (data: unknown): void => {
    if (!isRecord(data)) {
      console.error("bad data from /api/add: not a record", data);
      return;
    }

    // Not sure where this goes, so let's fetch again.
    const auction = parseAuction(data.auction);
    if (auction !== undefined) {
      this.setState({page: "list", auctions: undefined});  // show loading

      fetch("/api/list").then(this.doListResp)
          .catch(() => this.doListError("failed to connect to server"));
    }
  };

  doAddError = (msg: string): void => {
    console.error(`Error fetching /api/add: ${msg}`);
  };
  
  doBidClick = (index: number, bidder: string, amount: number): void => {
    if (DEBUG) console.debug("add an auction and then show the full list");

    const args = {index, bidder, amount};
    fetch("/api/bid", {
        method: "POST", body: JSON.stringify(args),
        headers: {"Content-Type": "application/json"} })
      .then((res) => this.doBidResp(index, res))
      .catch(() => this.doBidError("failed to connect to server"));
  };

  doBidResp = (index: number, res: Response): void => {
    if (res.status === 200) {
      res.json().then((data) => this.doBidJson(index, data))
          .catch(() => this.doBidError("200 response is not JSON"));
    } else if (res.status === 400) {
      res.json().then(this.doBidError)
          .catch(() => this.doBidError("400 response is not text"));
    } else {
      this.doBidError(`bad status code from /api/bid: ${res.status}`);
    }
  };

  doBidJson = (index: number, data: unknown): void => {
    if (this.state.auctions === undefined)
      throw new Error("impossible");

    if (!isRecord(data)) {
      console.error("bad data from /api/bid: not a record", data);
      return;
    }

    const auction = parseAuction(data.auction);
    if (auction !== undefined) {
      // Re-display the details page with the new auction.
      const auctions = this.state.auctions.slice(0, index)
          .concat([auction])
          .concat(this.state.auctions.slice(index+1));
      this.setState({auctions});
    }
  };

  doBidError = (msg: string): void => {
    console.error(`Error fetching /api/bid: ${msg}`);
  };

  doRefreshAuctionClick = (index: number): void => {
    if (DEBUG) console.debug("add an auction and then show the full list");

    const args = {index};
    fetch("/api/get", {
        method: "POST", body: JSON.stringify(args),
        headers: {"Content-Type": "application/json"} })
      .then((res) => this.doGetResp(index, res))
      .catch(() => this.doGetError("failed to connect to server"));
  };

  doGetResp = (index: number, res: Response): void => {
    if (res.status === 200) {
      res.json().then((data) => this.doGetJson(index, data))
          .catch(() => this.doGetError("200 res is not JSON"));
    } else if (res.status === 400) {
      res.json().then(this.doGetError)
          .catch(() => this.doGetError("400 response is not text"));
    } else {
      this.doGetError(`bad status code from /api/refersh: ${res.status}`);
    }
  };

  doGetJson = (index: number, data: unknown): void => {
    if (this.state.auctions === undefined)
      throw new Error("impossible");

    if (!isRecord(data)) {
      console.error("bad data from /api/refresh: not a record", data);
      return;
    }

    const auction = parseAuction(data.auction);
    if (auction !== undefined) {
      // Re-display the details page with the new auction.
      const auctions = this.state.auctions.slice(0, index)
          .concat([auction])
          .concat(this.state.auctions.slice(index+1));
      this.setState({auctions});
    }
  };

  doGetError = (msg: string): void => {
    console.error(`Error fetching /api/refresh: ${msg}`);
  };
}